"""A utility iterator to handle default RedBrick pagination behavior."""

from typing import Any, Dict, List, Optional, Callable, Tuple


class PaginationIterator:
    """Construct Labelset Iterator."""

    def __init__(self, func: Callable[[Optional[str]], Tuple[Any, ...]]) -> None:
        """Construct LabelsetIterator."""
        self.cursor: Optional[str] = None
        self.datapoints_batch: Optional[List[Dict]] = None
        self.datapoints_batch_index: Optional[int] = None

        self.func = func

        self.total = 0

    def __iter__(self) -> Any:
        """Get iterator."""
        return self

    def __len__(self) -> int:
        """Get length of iteration."""
        return self.total

    def __next__(self) -> Dict:
        """Get next batch of labels / datapoint."""
        # If cursor is None and current datapoints_batch has been processed
        if (
            self.datapoints_batch_index is not None
            and self.cursor is None
            and self.datapoints_batch
            and len(self.datapoints_batch) == self.datapoints_batch_index
        ):
            raise StopIteration

        # If current datapoints_batch is None or we have finished
        # processing current datapoints_batch
        if (
            self.datapoints_batch is None
            or len(self.datapoints_batch) == self.datapoints_batch_index
        ):
            datapoints_batch: List[Dict]
            cursor: Optional[str]
            datapoints_batch, cursor, *_ = self.func(self.cursor)
            self.datapoints_batch = datapoints_batch
            self.cursor = cursor
            self.datapoints_batch_index = 0
            self.total += len(self.datapoints_batch)

        # Current entry to return
        if self.datapoints_batch and self.datapoints_batch_index is not None:
            entry = self.datapoints_batch[self.datapoints_batch_index]
            self.datapoints_batch_index += 1

            return entry

        raise StopIteration



"""Abstract interface to active learning API."""


from typing import Optional, List, Dict, Tuple

import aiohttp

from redbrick.common.client import RBClient
from redbrick.common.learning import (
    LearningControllerInterface,
    LearningController2Interface,
)

from .shards import TAXONOMY_SHARD


class LearningRepo(LearningControllerInterface):
    """Abstract interface to Active Learning APIs."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def check_for_job(
        self, org_id: str, project_id: str, stage_name: str
    ) -> Optional[int]:
        """Check for a job, returns cycle number."""
        query = """
        query  ($orgId:UUID!, $projectId:UUID!, $stageName: String!){
            activeLearningClientSummary(orgId:$orgId, projectId:$projectId, stageName: $stageName){
                availableJob {
                    cycleStatus
                    cycle
                }
            }
        }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
        }
        result = self.client.execute_query(query, variables)

        if result.get("activeLearningClientSummary", {}).get("availableJob"):
            return int(result["activeLearningClientSummary"]["availableJob"]["cycle"])

        return None

    def get_batch_of_tasks(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        first: int = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[dict], Optional[str]]:
        """Get batch of tasks, paginated."""
        query = """
        query  ($orgId:UUID!, $projectId:UUID!, $stageName: String!, $first: Int!, $cursor: String){
            activeLearningClientSummary(orgId:$orgId, projectId:$projectId, stageName: $stageName){
                tdType
                tasks (cursor: $cursor, first: $first){
                    cursor
                    entries {
                        taskId
                        datapoint {
                            name
                            itemsPresigned:items (presigned:true)
                            items(presigned:false)
                        }
                        groundTruth {
                            labelsData(interpolate: true) {
                                %s
                            }
                        }
                    }
                }
            }
        }
        """
        variables = {
            "cursor": cursor,
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "first": 100,
        }
        result = self.client.execute_query(query, variables)
        return (
            result["activeLearningClientSummary"]["tasks"]["entries"],
            result["activeLearningClientSummary"]["tasks"]["cursor"],
        )

    def get_taxonomy_and_type(
        self, org_id: str, project_id: str, stage_name: str
    ) -> Tuple[dict, str]:
        """Get the taxonomy for active learning."""
        query = (
            """
        query  ($orgId:UUID!, $projectId:UUID!, $stageName: String!){
            activeLearningClientSummary(orgId:$orgId, projectId:$projectId, stageName: $stageName){
                tdType
                taxonomy {
                    %s
                }
            }
        }
        """
            % TAXONOMY_SHARD
        )
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
        }
        result = self.client.execute_query(query, variables)

        return (
            result["activeLearningClientSummary"]["taxonomy"],
            result["activeLearningClientSummary"]["tdType"],
        )

    def send_batch_learning_results(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        cycle: int,
        tasks: List[Dict],
    ) -> None:
        """
        Send a batch of learning results.

        tasks is a list of dictionaries containing the following keys:
        {
            "taskId": "<>",
            "score": [0,1],
            "labels": [{ }]  // see standard label format
        }
        """
        raise NotImplementedError()

    async def send_batch_learning_results_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        cycle: int,
        tasks: List[Dict],
    ) -> None:
        """Perform send_batch_learning_results with asyncio."""
        query = """
        mutation(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String!
            $tasks: [AlTaskLabelsInput!]!
            $cycle: Int!
        ) {
            sendAlLabels(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                tasks: $tasks
                cycle: $cycle
            ) {
                ok
            }
        }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "cycle": cycle,
            "tasks": tasks,
        }
        await self.client.execute_query_async(aio_client, query, variables)

    def set_cycle_status(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        cycle: int,
        cycle_status: str,
    ) -> None:
        """Set status of current training cycle."""
        query = """
        mutation(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String!
            $cycle: Int!,
            $status: CycleStatus!
        ) {
            updateAlCycleStatus(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                cycle: $cycle
                cycleStatus: $status
            ) {
                ok
            }
        }
        """

        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "cycle": cycle,
            "status": cycle_status,
        }
        self.client.execute_query(query, variables)


class Learning2Repo(LearningController2Interface):
    """Abstract interface to Active Learning APIs."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def check_for_job(self, org_id: str, project_id: str) -> Dict:
        """Check for a job, and num new tasks."""
        query = """
        query checkForJob ($orgId:UUID!, $projectId: UUID!) {
            currentCycleTracker(orgId:$orgId, projectId:$projectId) {
                newTasks
                newTasksProcessing
                isProcessing
            }
        }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        result: Dict[str, Dict] = self.client.execute_query(query, variables)

        return result["currentCycleTracker"]

    def get_taxonomy_and_type(self, org_id: str, project_id: str) -> Tuple[dict, str]:
        """Get the taxonomy for active learning."""
        query = (
            """
        query projectTypeAndTaxonomy ($orgId: UUID!, $projectId: UUID!) {
            project(orgId: $orgId, projectId: $projectId){
                tdType
                taxonomy {
                    %s
                }
            }
        }

        """
            % TAXONOMY_SHARD
        )

        variables = {"orgId": org_id, "projectId": project_id}
        result: Dict = self.client.execute_query(query, variables)
        return (
            result["project"]["taxonomy"],
            result["project"]["tdType"],
        )

    async def update_prelabels_and_priorities(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        tasks: List[Dict],
    ) -> None:
        """
        Perform send_batch_learning_results with asyncio.

        tasks is a list of dictionaries containing the following keys:
        {
            "taskId": "<>",
            "priority": [0,1],
            "labels": [{ }]  // see standard label format
        }
        """
        query = """
        mutation updatePrelabelsAndPriorities(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String!
            $tasks: [TaskUpdateInput!]!
        ) {
            updatePrelabelsOrPriority(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                tasks: $tasks
            ) {
                ok
            }
        }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "tasks": tasks,
        }

        await self.client.execute_query_async(aio_client, query, variables)

    def start_processing(self, org_id: str, project_id: str) -> None:
        """Set status of current training cycle."""
        query = """
        mutation ($orgId:UUID!, $projectId:UUID!){
            startCycleProcessing(orgId:$orgId, projectId:$projectId){
                ok
            }
        }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        self.client.execute_query(query, variables)

    def end_processing(self, org_id: str, project_id: str) -> None:
        """Set status of current training cycle."""
        query = """
        mutation ($orgId:UUID!, $projectId:UUID!){
            endCycleProcessing(orgId:$orgId, projectId:$projectId){
                ok
            }
        }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        self.client.execute_query(query, variables)



"""Logging functions."""
import os
import sys
from typing import Callable, Union, Any, TypeVar, cast
import logging
from functools import wraps


class Formatter(logging.Formatter):
    """Custom formatter."""

    grey = "\x1b[38;22m"
    blue = "\x1b[34;22m"
    yellow = "\x1b[33;22m"
    red = "\x1b[31;22m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    format_color = "%(levelname)s - "
    format_rest = "%(message)s"

    FORMATS = {
        logging.DEBUG: grey + format_color + reset + format_rest,
        logging.INFO: blue + format_color + reset + format_rest,
        logging.WARNING: yellow + format_color + reset + format_rest,
        logging.ERROR: red + format_color + reset + format_rest,
        logging.CRITICAL: bold_red + format_color + reset + format_rest,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format method."""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class Logger:
    """Custom logger."""

    def __init__(self) -> None:
        """Construct logging object."""
        # Create a custom logger
        logger_ = logging.getLogger(__name__)
        logger_.setLevel(logging.INFO)

        # Create handlers
        console = logging.StreamHandler()

        # Create formatters and add it to handlers
        console.setFormatter(Formatter())

        # Add handlers to the logger
        logger_.addHandler(console)

        self.logger = logger_


logger = Logger().logger


def print_info(text: str) -> None:
    """Log general information."""
    logger.info(text)


def print_warning(text: str) -> None:
    """Log warnings."""
    logger.warning(text)


def print_error(text: Union[str, Exception]) -> None:
    """Log errors."""
    logger.error(text)


Func = TypeVar("Func", bound=Callable[..., Any])


def handle_exception(func: Func) -> Func:
    """Decorate generic exception handler.

    Catch and trace full exception if REDBRICK_SDK_DEBUG is set, else just print the error message.
    Specifically meant for all user facing methods.
    """

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except (KeyboardInterrupt, Exception) as error:  # pylint: disable=broad-except
            if os.environ.get("REDBRICK_SDK_DEBUG"):
                raise error
            print_error(
                "User interrupted" if isinstance(error, KeyboardInterrupt) else error
            )
            sys.exit(1)

    return cast(Func, wrap)



"""Constants."""


MAX_CONCURRENCY = 30
MAX_RETRY_ATTEMPTS = 5

DEFAULT_URL = "https://api.redbrickai.com"

ORG_API_HAS_CHANGED = (
    "this api has changed recently, try running help(redbrick.get_org)"
    + " or visiting https://redbrick-sdk.readthedocs.io/en/stable/#redbrick.get_org"
)

PROJECT_API_HAS_CHANGED = (
    "this api has changed recently, try running help(redbrick.get_project)"
    + " or visiting https://redbrick-sdk.readthedocs.io/en/stable/#redbrick.get_project"
)



# pylint: disable=invalid-name
# noqa

"""
aioimgspy is a port of imgspy by Derek Lukacs for RedBrick AI.

======

imgspy finds the metadata (type, size) of an image given its url by fetching
as little as needed. This is a python implementation of `fastimage`_. Supports
image types BMP, CUR, GIF, ICO, JPEG, PNG, PSD, TIFF, WEBP.

.. _fastimage: https://github.com/sdsykes/fastimage

usage
-----

::
    >>> async with aiohttp.ClientSession() as session:
    ...     async with session.get(url) as response:
    ...         await aioimgspy.probe(response.content)

======

ORIGINAL license:

Copyright (c) 2017 Nazar Kanaev (nkanaev@live.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""


import struct


# All of this code is taken from imgspy, just added await statements when
# reading from stream.

# pylint: disable=too-many-statements,
# pylint: disable=too-many-branches, too-many-locals, too-many-return-statements
async def probe(stream):  # type: ignore
    """Probe for img dimensions."""
    w, h = None, None
    chunk = await stream.readexactly(26)

    if chunk.startswith(b"\x89PNG\r\n\x1a\n"):
        if chunk[12:16] == b"IHDR":
            w, h = struct.unpack(">LL", chunk[16:24])
        elif chunk[12:16] == b"CgBI":
            # fried png http://www.jongware.com/pngdefry.html
            chunk += await stream.readexactly(40 - len(chunk))
            w, h = struct.unpack(">LL", chunk[32:40])
        else:
            w, h = struct.unpack(">LL", chunk[8:16])
        return {"type": "png", "width": w, "height": h}
    if chunk.startswith(b"GIF89a") or chunk.startswith(b"GIF87a"):
        w, h = struct.unpack("<HH", chunk[6:10])
        return {"type": "gif", "width": w, "height": h}
    if chunk.startswith(b"\xff\xd8"):
        start = 2
        data = chunk
        while True:
            if data[start : start + 1] != b"\xff":
                return
            if data[start + 1] in b"\xc0\xc2":
                h, w = struct.unpack(">HH", data[start + 5 : start + 9])
                return {"type": "jpg", "width": w, "height": h}
            (segment_size,) = struct.unpack(">H", data[start + 2 : start + 4])
            data += await stream.readexactly(segment_size + 9)
            start = start + segment_size + 2
    if chunk.startswith(b"\x00\x00\x01\x00") or chunk.startswith(b"\x00\x00\x02\x00"):
        img_type = "ico" if chunk[2:3] == b"\x01" else "cur"
        num_images = struct.unpack("<H", chunk[4:6])[0]
        w, h = struct.unpack("BB", chunk[6:8])
        w = 256 if w == 0 else w
        h = 256 if h == 0 else h
        return {"type": img_type, "width": w, "height": h, "num_images": num_images}
    if chunk.startswith(b"BM"):
        headersize = struct.unpack("<I", chunk[14:18])[0]
        if headersize == 12:
            w, h = struct.unpack("<HH", chunk[18:22])
        elif headersize >= 40:
            w, h = struct.unpack("<ii", chunk[18:26])
        else:
            return
        return {"type": "bmp", "width": w, "height": h}
    if chunk.startswith(b"MM\x00\x2a") or chunk.startswith(b"II\x2a\x00"):
        w, h, orientation = None, None, None

        endian = ">" if chunk[0:2] == b"MM" else "<"
        offset = struct.unpack(endian + "I", chunk[4:8])[0]
        chunk += await stream.readexactly(offset - len(chunk) + 2)

        tag_count = struct.unpack(endian + "H", chunk[offset : offset + 2])[0]
        offset += 2
        for _ in range(tag_count):
            if len(chunk) - offset < 12:
                chunk += stream.readexactly(12)
            type_ = struct.unpack(endian + "H", chunk[offset : offset + 2])[0]
            data = struct.unpack(endian + "H", chunk[offset + 8 : offset + 10])[0]
            offset += 12
            if type_ == 0x100:
                w = data
            elif type_ == 0x101:
                h = data
            elif type_ == 0x112:
                orientation = data
            if all([w, h, orientation]):
                break

        if orientation >= 5:  # type: ignore
            w, h = h, w
        return {"type": "tiff", "width": w, "height": h, "orientation": orientation}
    if chunk[:4] == b"RIFF" and chunk[8:15] == b"WEBPVP8":
        w, h = None, None
        type_ = chunk[15:16]
        chunk += await stream.readexactly(30 - len(chunk))
        if type_ == b" ":
            w, h = struct.unpack("<HH", chunk[26:30])
            w, h = w & 0x3FFF, h & 0x3FFF
        elif type_ == b"L":
            w = 1 + (((ord(chunk[22:23]) & 0x3F) << 8) | ord(chunk[21:22]))
            h = 1 + (
                ((ord(chunk[24:25]) & 0xF) << 10)
                | (ord(chunk[23:24]) << 2)
                | ((ord(chunk[22:23]) & 0xC0) >> 6)
            )
        elif type_ == b"X":
            w = 1 + struct.unpack("<I", chunk[24:27] + b"\x00")[0]
            h = 1 + struct.unpack("<I", chunk[27:30] + b"\x00")[0]
        return {"type": "webp", "width": w, "height": h}
    if chunk.startswith(b"8BPS"):
        h, w = struct.unpack(">LL", chunk[14:22])
        return {"type": "psd", "width": w, "height": h}



"""Abstract interface to Labeling APIs."""

from typing import Optional, List, Dict, Tuple
import aiohttp

from redbrick.common.client import RBClient
from redbrick.common.labeling import LabelingControllerInterface


class LabelingRepo(LabelingControllerInterface):
    """Implementation of manual labeling apis."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def get_labeling_tasks(
        self, org_id: str, project_id: str, stage_name: str, count: int = 5
    ) -> List[Dict]:
        """Get labeling tasks."""
        query = """
        mutation assignLabelingTasks(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String!
            $count: Int!
        )  {
                assignLabelingTasks(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                count: $count
                ) {
                    orgId
                    projectId
                    stageName
                    state
                    taskId
                    taskType
                    completionTimeMs
                    progressSavedAt
                    assignedTo {
                        userId
                        loggedInUser
                    }
                    taxonomy {
                        name
                        version
                    }
                    datapoint {
                        itemsPresigned: items(presigned: true)
                        items(presigned: false)
                        dataType
                        name
                    }
                    taskData {
                        subName
                        taskType
                        createdAt
                        createdBy
                        labelsData(interpolate: true)
                }
            }
        }
        """

        response = self.client.execute_query(
            query,
            {
                "orgId": org_id,
                "projectId": project_id,
                "stageName": stage_name,
                "count": count,
            },
        )
        tasks: List[Dict] = response["assignLabelingTasks"]
        return tasks

    async def presign_labels_path(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        task_id: str,
        file_type: str,
    ) -> Dict:
        """Presign labels path."""
        query = """
        query presignLabelsPathSDK(
            $orgId: UUID!
            $projectId: UUID!
            $taskId: UUID!
            $fileType: String!
        ) {
            presignLabelsPath(
                orgId: $orgId
                projectId: $projectId
                taskId: $taskId
                fileType: $fileType
            ) {
                fileName
                filePath
                presignedUrl
            }
        }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskId": task_id,
            "fileType": file_type,
        }
        response = await self.client.execute_query_async(session, query, variables)
        presigned: Dict = response["presignLabelsPath"]
        return presigned

    async def put_labeling_results(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        labels_data: str,
        labels_path: Optional[str] = None,
        finished: bool = True,
    ) -> None:
        """Put Labeling results."""
        query = """
        mutation putTaskAndLabels(
        $orgId: UUID!
        $projectId: UUID!
        $stageName: String!
        $taskId: UUID!
        $elapsedTimeMs: Int!
        $finished: Boolean!
        $labelsData: String
        $labelsPath: String
        ) {
            putManualLabelingTaskAndLabels(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                taskId: $taskId
                elapsedTimeMs: $elapsedTimeMs
                finished: $finished
                labelsData: $labelsData
                labelsPath: $labelsPath
            ) {
                ok
            }
        }
        """

        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskId": task_id,
            "labelsData": labels_data,
            "labelsPath": labels_path,
            "finished": finished,
            "elapsedTimeMs": 0,
        }
        await self.client.execute_query_async(session, query, variables)

    async def put_review_task_result(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: str,
        review_val: bool,
    ) -> None:
        """Put review result for task."""
        query = """
        mutation putReviewTask(
        $orgId: UUID!
        $projectId: UUID!
        $stageName: String!
        $reviewVal: Boolean!
        $taskId: UUID!
        $elapsedTimeMs: Int!
        ) {
            putExpertReviewTask(
            orgId: $orgId
            projectId: $projectId
            stageName: $stageName
            reviewVal: $reviewVal
            taskId: $taskId
            elapsedTimeMs: $elapsedTimeMs
            ) {
            ok
            }
        }
        """

        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskId": task_id,
            "reviewVal": review_val,
            "elapsedTimeMs": 0,
        }

        await self.client.execute_query_async(session, query, variables)

    def assign_task(
        self, org_id: str, project_id: str, stage_name: str, task_id: str, email: str
    ) -> None:
        """Assign task to specified email."""
        query_string = """
        mutation assignTaskByEmailSDK(
            $orgId: UUID!,
            $projectId: UUID!,
            $stageName: String!,
            $taskId: UUID!,
            $email: String!
        ){
            assignTask(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                taskId: $taskId
                email: $email
            ) {
                task {
                    taskId
                }
            }
        }
        """

        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskId": task_id,
            "email": email,
        }

        self.client.execute_query(query_string, query_variables)

    def get_tasks_queue(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        first: int,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], str]:
        """Get task queue."""
        query_string = """
            query(
                $orgId: UUID!
                $projectId: UUID!
                $stageName: String!
                $first: Int!
                $after: String
                ) {
                taskQueue(
                    orgId: $orgId
                    projectId: $projectId
                    stageName: $stageName
                    first: $first
                    after: $after
                ) {
                    cursor
                    entries {
                        taskId
                        state
                        assignedTo {
                            email
                            userId
                        }
                        datapoint {
                            itemsPresigned: items(presigned: true)
                            items(presigned: false)
                            dataType
                            name
                    }
                    }
                }
            }

        """

        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "first": first,
            "after": cursor,
        }

        response = self.client.execute_query(query_string, query_variables)
        return response["taskQueue"]["entries"], response["taskQueue"]["cursor"]

    def get_task_queue_count(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        for_active_learning: bool = False,
    ) -> int:
        """Get the length of the task queue for showing loading."""
        query_string = """
            query ($orgId: UUID!
                $projectId: UUID!
                $stageName: String!)  {
                manualLabelingStageSummary(
                    orgId:$orgId,
                    projectId:$projectId,
                    stageName:$stageName
                ){
                    taskStatusSummary {
                    assignedCount
                    unassignedCount
                    inProgressCount
                    }
                }
            }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
        }

        response = self.client.execute_query(query_string, query_variables)
        summary = response["manualLabelingStageSummary"]["taskStatusSummary"]
        return (
            int(summary["unassignedCount"])
            if for_active_learning
            else int(
                summary["assignedCount"]
                + summary["unassignedCount"]
                + summary["inProgressCount"]
            )
        )

    async def move_task_to_start(
        self,
        session: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        task_id: str,
    ) -> None:
        """Move groundtruth task back to start."""
        query = """
        mutation moveTaskToStart(
            $orgId: UUID!
            $projectId: UUID!
            $taskId: UUID!
        ) {
            moveTaskToStart(
                orgId: $orgId
                projectId: $projectId
                taskId: $taskId
            ) {
                ok
            }
        }
        """

        variables = {"orgId": org_id, "projectId": project_id, "taskId": task_id}

        await self.client.execute_query_async(session, query, variables)



"""CLI init command."""
import os
from argparse import ArgumentParser, Namespace

from halo.halo import Halo  # type: ignore

from redbrick.cli.input import CLIInputNumber, CLIInputSelect, CLIInputText
from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIInitInterface
from redbrick.common.enums import LabelType
from redbrick.organization import RBOrganization


class CLIInitController(CLIInitInterface):
    """CLI init command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize init sub commands."""
        parser.add_argument("--name", "-n", help="Project name")
        parser.add_argument("--label", "-l", help="Label type")
        parser.add_argument("--taxonomy", "-t", help="Taxonomy name")
        parser.add_argument("--reviews", "-r", help="Number of review stages")
        parser.add_argument(
            "path", nargs="?", default=".", help="Local path of project"
        )

    def handler(self, args: Namespace) -> None:
        """Handle init command."""
        self.args = args
        project = CLIProject.from_path(path=self.args.path, required=False)
        assert project is None, f"Already a RedBrick project {project.path}"

        path = os.path.realpath(self.args.path)
        if os.path.exists(path):
            if not os.path.isdir(path):
                raise Exception(f"Not a directory {path}")
            if os.listdir(path):
                raise Exception(f"{path} is not empty")
        else:
            os.makedirs(path)

        self.project = CLIProject(path=self.args.path, required=False)

        self.handle_init()

    def handle_init(self) -> None:
        """Handle empty sub command."""
        assert self.project.creds.exists, "Credentials missing"

        with Halo(text="Fetching organization", spinner="dots") as spinner:
            try:
                org = RBOrganization(self.project.context, self.project.creds.org_id)
                spinner.succeed(str(org))
            except Exception as error:
                spinner.fail()
                raise error

        with Halo(text="Fetching taxonomies", spinner="dots") as spinner:
            try:
                taxonomies = org.taxonomies()
                spinner.succeed()
            except Exception as error:
                spinner.fail()
                raise error

        name = CLIInputText(
            self.args.name, "Name", os.path.basename(self.project.path)
        ).get()
        label_type = LabelType(
            CLIInputSelect(
                self.args.label, "Label", [label.value for label in LabelType]
            ).get()
        )
        taxonomy = CLIInputSelect(self.args.taxonomy, "Taxonomy", taxonomies).get()
        reviews = int(CLIInputNumber(self.args.reviews, "Reviews").get())

        with Halo(text="Creating project", spinner="dots") as spinner:
            try:
                project = org.create_project(name, label_type, taxonomy, reviews)
                spinner.succeed(str(project))
            except Exception as error:
                spinner.fail()
                raise error

        self.project.initialize_project(org, project)



"""Container for low-level methods to communicate with API."""


class RBContext:
    """Basic context for accessing low level functionality."""

    def __init__(self, api_key: str, url: str) -> None:
        """Construct RedBrick client singleton."""
        # pylint: disable=import-outside-toplevel
        from .client import RBClient
        from .export import ExportControllerInterface
        from .upload import UploadControllerInterface
        from .labeling import LabelingControllerInterface
        from .learning import LearningControllerInterface, LearningController2Interface
        from .project import ProjectRepoInterface

        self.client = RBClient(api_key=api_key, url=url)

        self.export: ExportControllerInterface
        self.upload: UploadControllerInterface
        self.labeling: LabelingControllerInterface
        self.learning: LearningControllerInterface
        self.learning2: LearningController2Interface
        self.project: ProjectRepoInterface

    def __str__(self) -> str:
        """Get string representation."""
        return repr(self) + (
            "***" + self.client.api_key[-3:] if self.client.api_key else ""
        )



"""CLI export command."""
import asyncio
import os
import re
import json
from datetime import datetime, timezone
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import List, Dict
import shutil

from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIExportInterface
from redbrick.coco.coco_main import _get_image_dimension_map, coco_converter
from redbrick.common.enums import LabelType
from redbrick.utils.files import uniquify_path, download_files
from redbrick.utils.logging import print_info, print_warning, print_error


class CLIExportController(CLIExportInterface):
    """CLI export command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize export sub commands."""
        parser.add_argument(
            "type",
            nargs="?",
            default=self.TYPE_LATEST,
            help=f"Export type: ({self.TYPE_LATEST} [default], {self.TYPE_GROUNDTRUTH}, <task id>)",
        )
        parser.add_argument(
            "--format",
            "-f",
            choices=[
                self.FORMAT_REDBRICK,
                self.FORMAT_COCO,
                self.FORMAT_MASK,
                self.FORMAT_NIFTI,
            ],
            help="Export format (Default: "
            + f"{self.FORMAT_MASK} if project type is {LabelType.IMAGE_SEGMENTATION}, "
            + f"{self.FORMAT_NIFTI} if project type is {LabelType.DICOM_SEGMENTATION}, "
            + f"else {self.FORMAT_REDBRICK})",
        )
        parser.add_argument(
            "--with-files",
            action="store_true",
            help="Export with files (e.g. images/video frames)",
        )
        parser.add_argument(
            "--clear-cache", action="store_true", help="Clear local cache"
        )
        parser.add_argument(
            "--concurrency",
            "-c",
            type=int,
            default=10,
            help="Concurrency value (Default: 10)",
        )
        parser.add_argument(
            "--fill-holes",
            action="store_true",
            help=f"Fill holes (for {self.FORMAT_MASK} export format)",
        )
        parser.add_argument(
            "--max-hole-size",
            type=int,
            default=30,
            help=f"Max hole size (for {self.FORMAT_MASK} export format. Default: 30)",
        )
        parser.add_argument(
            "--destination",
            "-d",
            default=".",
            help="Destination directory (Default: current directory)",
        )

    def handler(self, args: Namespace) -> None:
        """Handle export command."""
        self.args = args
        project = CLIProject.from_path()
        assert project, "Not a valid project"
        self.project = project

        self.handle_export()

    def handle_export(self) -> None:
        """Handle empty sub command."""
        # pylint: disable=too-many-statements, too-many-locals, too-many-branches, protected-access

        from redbrick.export.public import (  # pylint: disable=import-outside-toplevel
            Export,
        )

        format_type = (
            self.args.format
            if self.args.format
            else (
                self.FORMAT_MASK
                if self.project.project.project_type == LabelType.IMAGE_SEGMENTATION
                else self.FORMAT_NIFTI
                if self.project.project.project_type == LabelType.DICOM_SEGMENTATION
                else self.FORMAT_REDBRICK
            )
        )
        if (
            format_type == self.FORMAT_MASK
            and self.project.project.project_type != LabelType.IMAGE_SEGMENTATION
        ):
            raise Exception(
                f"{self.FORMAT_MASK} is available for {LabelType.IMAGE_SEGMENTATION} projects."
            )
        if (
            format_type == self.FORMAT_NIFTI
            and self.project.project.project_type != LabelType.DICOM_SEGMENTATION
        ):
            raise Exception(
                f"{self.FORMAT_NIFTI} is available for {LabelType.DICOM_SEGMENTATION} projects."
            )
        if (
            self.args.type not in (self.TYPE_LATEST, self.TYPE_GROUNDTRUTH)
            and re.match(
                r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$",
                self.args.type.strip().lower(),
            )
            is None
        ):
            raise ArgumentError(None, "")

        if self.args.clear_cache or self.args.with_files:
            self.project.cache.clear_cache(True)

        cached_datapoints: Dict = {}
        cache_timestamp = None
        dp_conf = self.project.conf.get_section("datapoints")
        if dp_conf and "timestamp" in dp_conf:
            cached_dp = self.project.cache.get_data("datapoints", dp_conf["cache"])
            if isinstance(cached_dp, dict):
                cached_datapoints = cached_dp
                cache_timestamp = int(dp_conf["timestamp"])

        cached_dimensions: Dict = {}
        dim_cache = self.project.conf.get_option("dimensions", "cache")
        cached_dim = self.project.cache.get_data("dimensions", dim_cache)
        if isinstance(cached_dim, dict):
            cached_dimensions = cached_dim

        current_timestamp = int(datetime.now(timezone.utc).timestamp())
        datapoints, taxonomy = self.project.project.export._get_raw_data_latest(
            self.args.concurrency, False, cache_timestamp
        )

        print_info(f"Refreshed {len(datapoints)} newly updated tasks")

        for datapoint in datapoints:
            task_id = datapoint["taskId"]
            if task_id not in cached_dimensions:
                cached_dimensions[task_id] = None

            cached_datapoints[task_id] = datapoint

        dp_hash = self.project.cache.set_data("datapoints", cached_datapoints)
        self.project.conf.set_section(
            "datapoints", {"timestamp": str(current_timestamp), "cache": dp_hash}
        )
        self.project.conf.save()

        data = list(cached_datapoints.values())

        if self.args.type == self.TYPE_LATEST:
            pass
        elif self.args.type == self.TYPE_GROUNDTRUTH:
            data = [task for task in data if task["currentStageName"] == "END"]
        else:
            task_id = self.args.type.strip().lower()
            task = None
            for dpoint in data:
                if dpoint["taskId"] == task_id:
                    task = dpoint
                    break
            data = [task] if task else []

        loop = asyncio.get_event_loop()
        if format_type == self.FORMAT_COCO:
            compute_dims = [
                task
                for task in data
                if task["taskId"] not in cached_dimensions
                or cached_dimensions[task["taskId"]] is None
            ]
            if compute_dims:
                image_dimensions = loop.run_until_complete(
                    _get_image_dimension_map(compute_dims)
                )
                for task_id, dimension in image_dimensions.items():
                    cached_dimensions[task_id] = dimension

        dim_hash = self.project.cache.set_data("dimensions", cached_dimensions)
        self.project.conf.set_section("dimensions", {"cache": dim_hash})
        self.project.conf.save()

        export_dir = self.args.destination

        os.makedirs(export_dir, exist_ok=True)

        cli_data = [
            {key: value for key, value in task.items() if key != "itemsPresigned"}
            for task in data
        ]
        if format_type == self.FORMAT_MASK:
            mask_dir = os.path.join(export_dir, "masks")
            class_map = os.path.join(export_dir, "class_map.json")
            datapoint_map = os.path.join(export_dir, "datapoint_map.json")
            shutil.rmtree(mask_dir, ignore_errors=True)
            os.makedirs(mask_dir, exist_ok=True)
            Export._export_png_mask_data(
                cli_data,
                taxonomy,
                mask_dir,
                class_map,
                datapoint_map,
                self.args.fill_holes,
                self.args.max_hole_size,
            )
            print_info(f"Exported: {class_map}")
            print_info(f"Exported: {datapoint_map}")
            print_info(f"Exported masks to: {mask_dir}")
        elif format_type == self.FORMAT_NIFTI:
            nifti_dir = os.path.join(export_dir, "nifti")
            shutil.rmtree(nifti_dir, ignore_errors=True)
            os.makedirs(nifti_dir, exist_ok=True)
            task_map = os.path.join(export_dir, "tasks.json")
            Export._export_nifti_label_data(cli_data, nifti_dir, task_map)
            print_info(f"Exported: {task_map}")
            print_info(f"Exported nifti to: {nifti_dir}")
        else:
            export_path = os.path.join(
                export_dir,
                f"export_{format_type}_{self.args.type}_"
                + datetime.strftime(
                    datetime.fromtimestamp(current_timestamp), "%Y-%m-%d_%H-%M-%S"
                )
                + ".json",
            )

            cleaned_data = [
                {
                    key: value
                    for key, value in task.items()
                    if key not in ("currentStageName", "labelsPath")
                }
                for task in cli_data
            ]
            output = (
                coco_converter(cleaned_data, taxonomy, cached_dimensions)
                if format_type == self.FORMAT_COCO
                else cleaned_data
            )
            with open(export_path, "w", encoding="utf-8") as file_:
                json.dump(output, file_, indent=2)

            print_info(f"Exported successfully to: {export_path}")

        if self.args.with_files:
            try:
                loop.run_until_complete(self._download_files(data, export_dir))
            except Exception:  # pylint: disable=broad-except
                print_error("Failed to download files")

    async def _download_files(self, data: List[Dict], export_dir: str) -> None:
        project_type = str(self.project.project.project_type.value)
        if project_type.startswith("IMAGE_"):
            files_dir = os.path.join(export_dir, "images")
        elif project_type.startswith("VIDEO_"):
            files_dir = os.path.join(export_dir, "videos")
        elif project_type.startswith("DICOM_"):
            files_dir = os.path.join(export_dir, "dicom")
        else:
            print_warning(
                "Project data type needs to be IMAGE, VIDEO or DICOM to export files"
            )
            return

        shutil.rmtree(files_dir, ignore_errors=True)
        os.makedirs(files_dir, exist_ok=True)

        files = []
        for task in data:
            if project_type.startswith("IMAGE_"):
                task_dir = files_dir
                items = [
                    re.sub(r"[^\w.]+", "-", os.path.basename(task.get("name", "")))
                    or task["taskId"]
                ]
                fill_index = 0
            else:
                task_dir = uniquify_path(
                    os.path.join(
                        files_dir,
                        re.sub(
                            r"\W+",
                            "-",
                            os.path.basename(task.get("name", "")).split(".", 1)[0],
                        )
                        or task["taskId"],
                    )
                )
                os.makedirs(task_dir, exist_ok=True)
                items = [
                    re.sub(r"[^\w.]+", "-", os.path.basename(item))
                    for item in task["items"]
                ]
                fill_index = (
                    0
                    if all(items) and len(items) == len(set(items))
                    else len(str(len(items)))
                )

            for index, (item, url) in enumerate(zip(items, task["itemsPresigned"])):
                file_name, file_ext = os.path.splitext(item)
                files.append(
                    (
                        url,
                        os.path.join(
                            task_dir,
                            file_name
                            + (
                                ("-" + str(index).zfill(fill_index))
                                if fill_index
                                else ""
                            )
                            + file_ext,
                        ),
                    )
                )

        await download_files(files)



"""Public interface to labeling module."""
import os
import json
import asyncio
from typing import List, Dict, Optional
from copy import deepcopy
from functools import partial

import aiohttp
import tqdm

from redbrick.common.context import RBContext
from redbrick.common.constants import MAX_CONCURRENCY
from redbrick.utils.logging import print_error, print_info, handle_exception
from redbrick.utils.pagination import PaginationIterator
from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils.rb_label_utils import clean_rb_label
from redbrick.utils.files import NIFTI_FILE_TYPES, upload_files


class Labeling:
    """
    Perform programmatic labeling and review tasks.

    The Labeling class allows you to programmatically submit tasks.
    This can be useful for times when you want to make bulk actions
    e.g accepting several tasks, or make automated actions like using automated
    methods for review.
    """

    def __init__(
        self,
        context: RBContext,
        org_id: str,
        project_id: str,
        review: bool = False,
    ) -> None:
        """Construct Labeling."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id
        self.review = review

    @handle_exception
    def get_tasks(self, stage_name: str, count: int = 1) -> List[Dict]:
        """
        Get a list of tasks from stage_name.

        >>> project = redbrick.get_project(...)
        >>> label_tasks = project.labeling.get_tasks(...)
        >>> review_tasks = project.review.get_tasks...)

        Parameters
        -------------
        stage_name: str
            The name of the stage your want to get tasks from. You can
            find the stage name on the workflow overview on the project
            dashboard.

        count: int = 1
            The number of tasks to retrieve. We recommend keeping this
            < 50.

        Returns
        ----------
        List[Dict]
            Tasks that are queued in this stage. Please see reference doc
            for formats.
            https://docs.redbrickai.com/python-sdk/reference#task-objects
        """
        tasks = self.context.labeling.get_labeling_tasks(
            self.org_id, self.project_id, stage_name, count=count
        )

        def _clean_tasks(task: Dict) -> Dict:
            task_id = task["taskId"]
            labels = json.loads(task.get("taskData", {}).get("labelsData", "[]"))
            labels_cleaned = [clean_rb_label(label) for label in labels]

            items = task["datapoint"]["items"]
            items_presigned = task["datapoint"]["itemsPresigned"]
            name = task["datapoint"]["name"]

            return {
                "taskId": task_id,
                "labels": labels_cleaned,
                "items": items,
                "itemsPresigned": items_presigned,
                "name": name,
            }

        return [_clean_tasks(task) for task in tasks]

    async def _put_task(
        self, session: aiohttp.ClientSession, stage_name: str, task: Dict
    ) -> Optional[Dict]:
        task_id = task["taskId"]
        try:
            if self.review:
                review_val = task["reviewVal"]
                await self.context.labeling.put_review_task_result(
                    session,
                    self.org_id,
                    self.project_id,
                    stage_name,
                    task_id,
                    review_val,
                )
            else:
                labels_path: Optional[str] = None
                if (
                    task.get("labelsPath")
                    and str(task["labelsPath"]).endswith(".nii")
                    and os.path.isfile(task["labelsPath"])
                ):
                    file_type = NIFTI_FILE_TYPES["nii"]
                    presigned = await self.context.labeling.presign_labels_path(
                        session, self.org_id, self.project_id, task_id, file_type
                    )
                    if (
                        await upload_files(
                            [
                                (
                                    task["labelsPath"],
                                    presigned["presignedUrl"],
                                    file_type,
                                )
                            ],
                            "Uploading labels",
                            False,
                        )
                    )[0]:
                        labels_path = presigned["filePath"]

                await self.context.labeling.put_labeling_results(
                    session,
                    self.org_id,
                    self.project_id,
                    stage_name,
                    task_id,
                    json.dumps(task["labels"]),
                    labels_path,
                    not bool(task.get("draft")),
                )

        except ValueError as error:
            print_error(error)
            point_error = deepcopy(task)
            point_error["error"] = error
            return point_error
        return None

    async def _put_tasks(self, stage_name: str, tasks: List[Dict]) -> List[Dict]:
        conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
        async with aiohttp.ClientSession(connector=conn) as session:
            coros = [self._put_task(session, stage_name, task) for task in tasks]
            temp = await gather_with_concurrency(10, coros, "Uploading tasks")
        await asyncio.sleep(0.250)  # give time to close ssl connections
        return [val for val in temp if val]

    @handle_exception
    def put_tasks(self, stage_name: str, tasks: List[Dict]) -> List[Dict]:
        """
        Put tasks with new labels or review result.

        >>> project = redbrick.get_project(...)
        >>> project.labeling.put_tasks(...)
        >>> project.review.put_tasks(...)

        Parameters
        --------------
        stage_name: str
            The stage to which you want to submit the tasks. This must be the
            same stage as which you called get_tasks on.

        tasks: List[Dict]
            Tasks with new labels or review result. Please see doc for format.
            https://docs.redbrickai.com/python-sdk/programmatically-label-and-review

        Returns
        ---------------
        List[Dict]
            A list of tasks that failed the upload.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._put_tasks(stage_name, tasks))

    @handle_exception
    def assign_task(self, stage_name: str, task_id: str, email: str) -> None:
        """Assign task to specified email."""
        self.context.labeling.assign_task(
            self.org_id, self.project_id, stage_name, task_id, email
        )

    @handle_exception
    def get_task_queue(self, stage_name: str, concurrency: int = 200) -> List[Dict]:
        """Get all tasks in queue."""
        temp = self.context.labeling.get_tasks_queue
        my_iter = PaginationIterator(
            partial(temp, self.org_id, self.project_id, stage_name, concurrency)
        )

        count = self.context.labeling.get_task_queue_count(
            self.org_id, self.project_id, stage_name
        )

        def _parse_entry(item: Dict) -> Dict:
            return {
                "taskId": item["taskId"],
                "assignedTo": (item.get("assignedTo") or {}).get("email"),
                "status": item["state"],
                "items": item["datapoint"]["items"],
                "itemsPresigned": item["datapoint"]["itemsPresigned"],
                "name": item["datapoint"]["name"],
            }

        print_info("Downloading tasks in stage")
        return [
            _parse_entry(val)
            for val in tqdm.tqdm(my_iter, unit=" datapoints", total=count)
        ]

    async def _tasks_to_start(self, task_ids: List[str]) -> None:
        conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
        async with aiohttp.ClientSession(connector=conn) as session:
            coros = [
                self.context.labeling.move_task_to_start(
                    session, self.org_id, self.project_id, task_id
                )
                for task_id in task_ids
            ]
            await gather_with_concurrency(10, coros, "Moving tasks to Start")
        await asyncio.sleep(0.250)  # give time to close ssl connections

    @handle_exception
    def move_tasks_to_start(self, task_ids: List[str]) -> None:
        """Move groundtruth tasks back to start."""
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._tasks_to_start(task_ids))



"""Handler for file upload/download."""
import os
import gzip
from typing import List, Optional, Tuple, Set

import asyncio
import aiohttp
import aiofiles
from yarl import URL
import tenacity
from natsort import natsorted, ns

from redbrick.common.constants import MAX_CONCURRENCY, MAX_RETRY_ATTEMPTS
from redbrick.utils.async_utils import gather_with_concurrency

IMAGE_FILE_TYPES = {
    "png": "image/png",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "bmp": "image/bmp",
}

VIDEO_FILE_TYPES = {
    "mp4": "video/mp4",
}

DICOM_FILE_TYPES = {
    "dcm": "application/dicom",
}

JSON_FILE_TYPES = {"json": "application/json"}

NIFTI_FILE_TYPES = {"nii": "application/octet-stream"}

FILE_TYPES = {
    **IMAGE_FILE_TYPES,
    **VIDEO_FILE_TYPES,
    **DICOM_FILE_TYPES,
    **JSON_FILE_TYPES,
    **NIFTI_FILE_TYPES,
}


def get_file_type(file_path: str) -> Tuple[str, str]:
    """
    Return file type.

    Return
    ------------
    [file_ext, file_type]
        file_ext: png, jpeg, jpg etc.
        file_type: this is the MIME file type e.g. image/png
    """
    file_ext = file_path.split("?", 1)[0].rstrip("/").rsplit(".", 1)[-1].lower()

    if file_ext not in FILE_TYPES:
        raise ValueError(
            f"Unsupported file type {file_ext}! Only {','.join(FILE_TYPES.keys())} are supported"
        )
    return file_ext, FILE_TYPES[file_ext]


def find_files_recursive(
    root: str, file_types: Set[str], multiple: bool = False
) -> List[List[str]]:
    """Find files recursively in a directory, that belong to a list of allowed file types."""
    if not os.path.isdir(root):
        return []

    items = []
    list_items = []
    discard_list_items = False

    for item in os.listdir(root):
        if item.startswith("."):
            continue
        path = os.path.join(root, item)
        if os.path.isdir(path):
            items.extend(find_files_recursive(path, file_types, multiple))
            discard_list_items = True
        elif os.path.isfile(path) and item.rsplit(".", 1)[-1].lower() in file_types:
            if multiple:
                if not discard_list_items:
                    list_items.append(path)
            else:
                items.append([path])
        else:
            discard_list_items = True

    if (
        not discard_list_items
        and len({item.rsplit(".", 1)[-1].lower() for item in list_items}) == 1
    ):
        items.append(list(natsorted(list_items, alg=ns.IGNORECASE)))  # type: ignore

    return items


def uniquify_path(path: str) -> str:
    """Provide unique path with number index."""
    filename, extension = os.path.splitext(path)
    counter = 1

    while os.path.exists(path):
        path = filename + " (" + str(counter) + ")" + extension
        counter += 1

    return path


async def upload_files(
    files: List[Tuple[str, str, str]],
    progress_bar_name: Optional[str] = "Uploading files",
    keep_progress_bar: bool = True,
) -> List[bool]:
    """Upload files from local path to url (file path, presigned url, file type)."""

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type((KeyboardInterrupt,)),
        retry_error_callback=lambda _: False,
    )
    async def _upload_file(
        session: aiohttp.ClientSession, path: str, url: str, file_type: str
    ) -> bool:
        if not path or not url or not file_type:
            return False
        async with aiofiles.open(path, mode="rb") as file_:  # type: ignore
            data = gzip.compress(await file_.read())

        async with session.put(
            url,
            headers={"Content-Type": file_type, "Content-Encoding": "gzip"},
            data=data,
        ) as response:
            if response.status == 200:
                return True
            raise ConnectionError(f"Error in uploading {path} to RedBrick")

    # limit to 30, default is 100, cleanup is done by session
    conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [
            _upload_file(session, path, url, file_type)
            for path, url, file_type in files
        ]
        uploaded = await gather_with_concurrency(
            10, coros, progress_bar_name, keep_progress_bar
        )

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return uploaded


async def download_files(
    files: List[Tuple[str, str]],
    progress_bar_name: Optional[str] = "Downloading files",
    keep_progress_bar: bool = True,
) -> List[Optional[str]]:
    """Download files from url to local path (presigned url, file path)."""

    @tenacity.retry(
        reraise=True,
        stop=tenacity.stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    )
    async def _download_file(
        session: aiohttp.ClientSession, url: str, path: str
    ) -> Optional[str]:
        # pylint: disable=no-member
        if not url or not path:
            return None
        async with session.get(URL(url, encoded=True)) as response:
            if response.status == 200:
                data = await response.read()
                if response.headers.get("Content-Encoding") == "gzip":
                    try:
                        data = gzip.decompress(data)
                    except Exception:  # pylint: disable=broad-except
                        pass
                path = uniquify_path(path)
                async with aiofiles.open(path, "wb") as file_:  # type: ignore
                    await file_.write(data)
                return path
            return None

    # limit to 30, default is 100, cleanup is done by session
    conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [_download_file(session, url, path) for url, path in files]
        paths = await gather_with_concurrency(
            10, coros, progress_bar_name, keep_progress_bar
        )

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return paths



"""Interface for getting basic information about a project."""
from datetime import datetime
from typing import Dict, List, Tuple, Optional
from abc import ABC, abstractmethod


class ProjectRepoInterface(ABC):
    """Abstract interface to Project APIs."""

    @abstractmethod
    def get_project(self, org_id: str, project_id: str) -> Dict:
        """
        Get project name and status.

        Raise an exception if project does not exist.
        """

    @abstractmethod
    def get_stages(self, org_id: str, project_id: str) -> List[Dict]:
        """Get stages."""

    @abstractmethod
    def create_project(
        self, org_id: str, name: str, stages: List[dict], td_type: str, tax_name: str
    ) -> Dict:
        """Create a project and return project_id."""

    @abstractmethod
    def get_org(self, org_id: str) -> Dict:
        """Get organization."""

    @abstractmethod
    def get_projects(self, org_id: str) -> List[Dict]:
        """Get all projects in organization."""

    @abstractmethod
    def get_taxonomies(self, org_id: str) -> List[Dict]:
        """Get a list of taxonomies."""

    @abstractmethod
    def delete_project(self, org_id: str, project_id: str) -> None:
        """Delete Project."""

    @abstractmethod
    def get_labeling_information(
        self,
        org_id: str,
        start_date: datetime,
        end_date: datetime,
        first: int,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get org labeling information."""
